local WidgetLib = TidyPlatesWidgets
local LocalVars = TidyPlatesNeonTankVariables
local theme = TidyPlatesThemeList["Neon/Tank"]
local valueToString = TidyPlatesUtility.abbrevNumber

local SelectionArt = "Interface\\Addons\\TidyPlates_Neon\\Media\\Neon_Select"
------------------------------
-- Auras to Track (experimental)
------------------------------
local TrackedAuras = {}
TrackedAuras["Necrotic Plague"] = {color = {r = .1,g = 1, b = 0, a = 1},}

------------------------------
-- Debuff and Roster Handler
------------------------------
local UnitLookupTable = {}
local DangerWatcher
	
-- Event Handler
local function DangerEvents(frame, event, unit)
	local isUnitInParty = UnitPlayerOrPetInParty(unit)
	local isUnitInRaid = UnitInRaid(unit)
	local isUnitPet = (unit == "pet")
	
	if isUnitInParty or isUnitInRaid or isUnitPet then
		local name = UnitName(unit)
		UnitLookupTable[name] = unit
		TidyPlates:Update()		-- Theoretically, this should not 'double-call' since it's queued
	end
end


local function CheckGroupMemberAggro(name)	
	local unitid = UnitLookupTable[name]
	if unitid then
		local unitaggro = UnitThreatSituation(unitid)
		if unitaggro and unitaggro > 1 then return true end
	end
end

local function CheckGroupMemberDebuffs(name)
	local unitid = UnitLookupTable[name]
	local debuffName, debuffType, debuffColor, debuffReference, debuffIndex, _
	local debuffPriority = 0
	
	if unitid then
		for debuffIndex = 1, 40 do
			debuffName, _, _, _, debuffType = UnitDebuff(unitid,debuffIndex)					-- Operational
			--debuffName, debuffType, _, _, _, _, _, _ = UnitBuff(unitid,debuffIndex)		-- For testing 

			if debuffName then
				debuffReference = TrackedAuras[debuffName] or TrackedAuras[debuffType]
				if debuffReference then
					debuffColor = debuffReference.color
					return debuffColor
				end
				
			end			
		end
	end
	return nil
end

------------------------------
-- Target Tracker
------------------------------

local TrackedUnits = {}
local TrackedUnitTargets = {}
local TrackedUnitTargetHistory = {}
local TargetWatcher

local function TargetWatcherEvents()
	local widget, plate
	local target, unitid, guid
	local changes = false
	wipe(TrackedUnits)
	
	-- Store target history
	for guid, target in pairs(TrackedUnitTargets) do
		TrackedUnitTargetHistory[guid] = target
		TrackedUnitTargets[guid] = nil
	end
	
	-- Reset the Tracking List
	for guid in pairs(TrackedUnits) do TrackedUnits[guid] = nil end
	
	-- Build a list of Trackable targets (via target, focus, and raid members)
	guid = UnitGUID("target")
	if guid then TrackedUnits[guid] = "target" end

	guid = UnitGUID("focus")
	if guid then TrackedUnits[guid] = "focus" end
	
	local raidsize = GetNumRaidMembers() - 1
	for index = 1, raidsize do
		unitid = "raid"..index.."target"
		guid = UnitGUID(unitid)
		if guid then TrackedUnits[guid] = unitid end
	end
	
	-- Build a list of the target's targets and check for changes
	for guid, unitid in pairs(TrackedUnits) do
		if unitid then 
			TrackedUnitTargets[guid] = UnitName(unitid.."target")
			if TrackedUnitTargets[guid] ~= TrackedUnitTargetHistory[guid] then changes = true end
		end
	end
	
	-- Call for indicator Update, if needed
	if changes then 
		TidyPlates:Update()			-- To Do: Make a better update hook: either update specific GUIDs or update only indicators
	end
end


---------------
-- Tank Monitor
---------------
local TankNames = {}
local TankWatcher

local function IsTankedByAnotherTank(unit)
	local targetOf
	if unit.guid then
		if unit.isTarget then targetOf = UnitName("targettarget")				-- Nameplate is a target
		elseif unit.isMouseover then targetOf = UnitName("mouseovertarget")		-- Nameplate is a mouseover
		else targetOf = TrackedUnitTargets[unit.guid] end
		
		if targetOf and TankNames[targetOf] then return true end
	end
	return false
end

local function TankWatcherEvents()
	local index, size
	if UnitInRaid("player") then
		size = GetNumRaidMembers() - 1
		for index = 1, size do
			local raidid = "raid"..tostring(index)
			
			local isAssigned = GetPartyAssignment("MAINTANK", raidid) or ("TANK" == UnitGroupRolesAssigned(raidid))
			
			if isAssigned then TankNames[UnitName(raidid)] = true 
			else TankNames[UnitName(raidid)] = nil end
		end			
	else 
		wipe(TankNames)
		if HasPetUI("player") then TankNames[UnitName("pet")] = true end			-- Adds your pet to the list
	end	
end

---------------
-- Theme Function Enabler
---------------
local NeonTankEnabled = false

local function EnableNeonTank()
	--[[
	Currently, this enables the watcher frames when Neon/Tank is activated.  It does NOT disable
	the watcher frames when the Theme is changed.  This is a known issue, and will be resolved 
	at some point in the future.  It doesn't greatly impact performance, which makes it low priority.
	--]]

	--[[	Experimental
	-- Aggro and Debuff Monitor
	DangerWatcher = CreateFrame("Frame")
	DangerWatcher:SetScript("OnEvent", DangerEvents)
	DangerWatcher:RegisterEvent("UNIT_AURA")
	DangerWatcher:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")
	--]]
	
	-- Target Tracker
	TargetWatcher = CreateFrame("Frame")
	TargetWatcher:RegisterEvent("PLAYER_REGEN_ENABLED")
	TargetWatcher:RegisterEvent("PLAYER_REGEN_DISABLED")
	TargetWatcher:RegisterEvent("PLAYER_TARGET_CHANGED")
	TargetWatcher:RegisterEvent("UNIT_THREAT_SITUATION_UPDATE")
	TargetWatcher:RegisterEvent("UNIT_TARGET")
	TargetWatcher:SetScript("OnEvent", TargetWatcherEvents)
	TargetWatcherEvents()
	
	-- Tank Monitor
	TankWatcher = CreateFrame("Frame")
	TankWatcher:RegisterEvent("RAID_ROSTER_UPDATE")
	TankWatcher:RegisterEvent("PARTY_MEMBERS_CHANGED")
	TankWatcher:RegisterEvent("PARTY_CONVERTED_TO_RAID")
	TankWatcher:SetScript("OnEvent", TankWatcherEvents)
	TankWatcherEvents()
	
	NeonTankEnabled = true
end
		


---------------
-- Text Delegates
---------------

local function SpellTextDelegate(unit)
	local spellname
	if unit.isCasting then 
		spellname = UnitCastingInfo("target") or UnitChannelInfo("target") or unit.spellName or ""
		return spellname
	else return "" end
end

local HealthTextFunctions = {
	--HEALTH_NONE
	function (health, healthmax) return "" end,
	--HEALTH_PCT
	function (health, healthmax) if health ~= healthmax then return "%"..ceil(100*(health/healthmax)) else return nil end end,
	--HEALTH_TOTAL
	function (health, healthmax) return valueToString(health) end,
	--HEALTH_DEF
	function (health, healthmax) if health ~= healthmax then return "-"..valueToString(healthmax - health) end end,
	--HEALTH_TOT_PCT
	function (health, healthmax) return valueToString(health).." / "..valueToString(healthmax).." ("..ceil(100*(health/healthmax)).."%)" end,
	--Health_Total_Pct_Def
	function (health, healthmax) return "+"..valueToString(health).." ("..ceil(100*(health/healthmax)).."%) -"..valueToString(healthmax - health) end,
}

local function HealthTextDelegate(unit)  
	return HealthTextFunctions[LocalVars.HealthText](unit.health, unit.healthmax, unit.targetOf)
end


---------------
-- Graphics Delegates
---------------
local function TankScale(unit)
	if InCombatLockdown() and unit.reaction == "HOSTILE" and  unit.threatSituation ~= "HIGH" and unit.type == "NPC"  then
		--if IsTankedByAnotherTank(unit) then return LocalVars.ScaleGeneral end
		if LocalVars.ScaleIgnoreNonElite and not unit.isElite then return LocalVars.ScaleGeneral end
		return LocalVars.ScaleLoose
	end
	return LocalVars.ScaleGeneral
end
	
local function TankAlpha(unit)
	if unit.isTarget then return 1
	else 	
		if unit.name == "Fanged Pit Viper" then return 0 end
		
		if LocalVars.OpacityHideNeutral and unit.reaction == "NEUTRAL" then return 0 end
		if LocalVars.OpacityHideNonElites and not unit.isElite then return 0 end
		if not UnitExists("target") then return 1 end
		return LocalVars.OpacityNonTarget, true
	end

end


-----------------
-- Combat Coloring
-----------------

local currentcolor
local function HealthColorDelegate(unit)
	if LocalVars.AggroHealth then
		if InCombatLockdown() and unit.reaction ~= "FRIENDLY" and unit.type == "NPC" then
				if unit.threatSituation ~= "HIGH" then 
					--if IsTankedByAnotherTank(unit) then currentcolor = LocalVars.AggroOtherColor 	-- Tank by another Raid Tank
					--else currentcolor = LocalVars.AggroLooseColor	end -- Not Tanked
					currentcolor = LocalVars.AggroLooseColor	-- Not Tanked
				else 	
					currentcolor = LocalVars.AggroTankedColor		-- Tanked By Me
				end
			return currentcolor.r, currentcolor.g, currentcolor.b
		end	
	end
	return unit.red, unit.green, unit.blue
end


local function ThreatColorDelegate(unit)

	--[[	Experimental
	if unit.reaction == "FRIENDLY" then
		--local color = CheckGroupMemberDebuffs(unit.name)		-- -- Do something!  Either return the color, or activate the debuff widget
		if CheckGroupMemberAggro(unit.name) then return 1, 0, 0, 1 end
		
		if color then
			return color.r, color.g, color.b, color.a
		end		
	end
	--]]

	if LocalVars.AggroBorder then
		if InCombatLockdown() and unit.reaction ~= "FRIENDLY" and unit.type == "NPC" then
			if unit.threatSituation ~= "HIGH" then 
				currentcolor = LocalVars.AggroLooseColor		-- Default/Simple
				--if IsTankedByAnotherTank(unit) then return 0, 0, 0, 0 end
				return currentcolor.r, currentcolor.g, currentcolor.b, 1
			end
		end
	end
	

	return 0, 0, 0, 0
end



---------------
-- Widgets
---------------

local DEBUFFMODE_ALL, DEBUFFMODE_FILTER = 1, 2

local function DebuffFilter(debuff)
	if LocalVars.WidgetDebuffMode == DEBUFFMODE_FILTER then
		if LocalVars.WidgetDebuffList[debuff.name] then return true end
	else return true end
	--if debuff.duration < 50  then return true end
end

local function OnInitializeDelegate(plate)
	-- if not NeonTankEnabled then EnableNeonTank() end

	-- Tug-o-Threat
	if LocalVars.WidgetTug and (not plate.widgets.WidgetTug) then 
			plate.widgets.WidgetTug = WidgetLib.CreateThreatLineWidget(plate)
			plate.widgets.WidgetTug:SetPoint("CENTER", plate, 0, 4)
			plate.widgets.WidgetTug._LowColor = LocalVars.TugWidgetLooseColor
			plate.widgets.WidgetTug._HighColor = LocalVars.TugWidgetAggroColor
			plate.widgets.WidgetTug._TankedColor = LocalVars.TugWidgetSafeColor
	end
	
	-- Target Selection Box
	if LocalVars.WidgetSelect and (not plate.widgets.SelectionBox) then 
			plate.widgets.SelectionBox = WidgetLib.CreateSelectionWidget(plate)
			plate.widgets.SelectionBox:SetWidgetTexture(SelectionArt)
			plate.widgets.SelectionBox:SetPoint("CENTER", 0, 0)
	end
	
	-- Short Debuffs
	if LocalVars.WidgetDebuff then
		if not plate.widgets.AuraIcon then
			plate.widgets.AuraIcon =  WidgetLib.CreateAuraWidget(plate)
			plate.widgets.AuraIcon:SetPoint("CENTER", plate, 15, 20)
			plate.widgets.AuraIcon:SetFrameLevel(plate:GetFrameLevel())
			-- For Filtering
			plate.widgets.AuraIcon.Filter = DebuffFilter
		end
	end
	
	-- Class 
	if LocalVars.WidgetClassIcon and (not plate.widgets.ClassWidget) then 
		plate.widgets.ClassWidget = WidgetLib.CreateClassWidget(plate)
		plate.widgets.ClassWidget:SetPoint("TOP", plate, 0, 3)
		plate.widgets.ClassWidget:SetScale(1.2)
	end
	
	-- Range
	if LocalVars.WidgetRange and (not plate.widgets.Range) then
			plate.widgets.Range = WidgetLib.CreateRangeWidget(plate)
			plate.widgets.Range:SetPoint("CENTER", 0, 0)
	end
	
	-- Combo Point Wheel
	if LocalVars.WidgetCombo then
		if not plate.widgets.WidgetCombo then 
			plate.widgets.WidgetCombo = WidgetLib.CreateComboPointWidget(plate)
			plate.widgets.WidgetCombo:SetPoint("CENTER", plate, 0, 10)
			plate.widgets.WidgetCombo:SetFrameLevel(plate:GetFrameLevel()+2)
		end
	end	

end

local RangeModeRef = { 9, 15, 28, 40 }

local function OnContextUpdateDelegate(plate, unit)

	-- Selection Box
	if LocalVars.WidgetSelect then plate.widgets.SelectionBox:Update(unit) end
	-- Combo Points
	if LocalVars.WidgetCombo then plate.widgets.WidgetCombo:UpdateContext(unit) end
	-- Tug-o-Threat
	if LocalVars.WidgetTug then plate.widgets.WidgetTug:UpdateContext(unit) end
end


local function OnUpdateDelegate(plate, unit)	
	-- Range Check
	if LocalVars.WidgetRange then plate.widgets.Range:Update(unit,RangeModeRef[LocalVars.RangeMode]) end
	-- Class Icon
	if LocalVars.WidgetClassIcon then plate.widgets.ClassWidget:Update(unit) end
	-- Short Debuffs
	if LocalVars.WidgetDebuff then plate.widgets.AuraIcon:Update(unit) 	end
end

local function StyleDelegate(unit)
	-- DefaultStyle, NoNameStyle, or NoHealthBarStyle
	if unit.reaction == "HOSTILE" then return "DefaultStyle"
	else return "NoNameStyle" end
end

---------------
-- Function Assignment - Tank Mode
---------------
theme.SetSpecialText = SpellTextDelegate
theme.SetSpecialText2 = HealthTextDelegate
theme.SetScale = TankScale
--theme.SetStyle = StyleDelegate
theme.SetAlpha = TankAlpha
theme.OnUpdate = OnUpdateDelegate
theme.OnInitialize = OnInitializeDelegate
theme.SetHealthbarColor = HealthColorDelegate
theme.SetThreatColor = ThreatColorDelegate
theme.OnContextUpdate = OnContextUpdateDelegate		-- 5.14




